__all__ = [
    "parallel_starmap",
    "starmap_with_kwargs",
    "apply_args_and_kwargs",
]

from itertools import repeat
from multiprocessing import pool as mp_pool
from typing import Any, Callable, Iterable, List, Mapping, Optional, Sequence, Type, TypeVar, Union

T = TypeVar("T")
U = TypeVar("U")

POOL = TypeVar("POOL", bound=mp_pool.Pool)


def starmap_with_kwargs(
    pool,
    fn: Callable,
    args_iter: Sequence[Iterable[Any]],
    kwargs_iter: Sequence[Mapping[str, Any]],
):
    args_for_starmap = zip(repeat(fn), args_iter, kwargs_iter)
    return pool.starmap(apply_args_and_kwargs, args_for_starmap)


def apply_args_and_kwargs(fn, args: Iterable[Any], kwargs: Mapping[str, Any]):
    return fn(*args, **kwargs)  # pragma: no cover


def _starmap_apply(
    fn: Callable[[List[Any]], U],
    args: Sequence[Any],
    kwargs: Mapping[str, Any],
) -> U:
    return fn(*args, **kwargs)  # type: ignore  # pragma: no cover


def parallel_starmap(
    callable: Callable[[Any], U],
    arguments: Sequence[T],
    keyword_arguments: Optional[Union[Sequence[Mapping[str, Any]], Mapping[str, Any]]] = None,
    pool_class: Optional[Type[mp_pool.Pool]] = None,
    processes: Optional[int] = None,
    chunk_size: Optional[int] = None,
    callback: Optional[Callable[[List[T]], Any]] = None,
    error_callback: Optional[Callable[[BaseException], None]] = None,
) -> List[U]:
    pool_class = pool_class or mp_pool.Pool
    with pool_class(processes=processes) as pool:
        starmap_arguments = zip(
            repeat(callable),
            arguments,
            (
                repeat(keyword_arguments or {})
                if not isinstance(keyword_arguments, Sequence)
                else keyword_arguments
            ),
        )

        async_results = [
            pool.starmap_async(
                _starmap_apply,
                (_,),
                chunksize=chunk_size,
                callback=callback,
                error_callback=error_callback,
            )
            for _ in starmap_arguments
        ]

        return [result for async_result in async_results for result in async_result.get()]



__all__ = [
    "DEFAULT_LOGGING_FORMAT",
    "get_logger",
    "get_formatter",
    "get_stdout_handler",
    "get_all_handlers",
    "check_formatter_equality",
    "check_handler_equality",
    "enable_stdout_logging",
    "LoggingMixin",
]

import logging
import sys

# Unfortunately, cannot import from decorators, because decorators imports from logging.
from functools import cache, cached_property
from typing import List, Optional, Union

DEFAULT_LOGGING_FORMAT = "%(name)s : %(levelname)s : %(asctime)s : %(message)s"


StrOrFormatter = Union[str, logging.Formatter]
StrOrLogger = Union[str, logging.Logger]
LogLevel = Union[str, int]


def get_logger(name: Optional[str] = None):
    return logging.getLogger(name=name)


@cache
def get_formatter(format: StrOrFormatter = DEFAULT_LOGGING_FORMAT) -> logging.Formatter:
    formatter = logging.Formatter(format) if isinstance(format, str) else format
    return formatter


@cache
def get_stdout_handler(
    format: StrOrFormatter = DEFAULT_LOGGING_FORMAT, level: LogLevel = "INFO"
) -> logging.StreamHandler:
    formatter = get_formatter(format)

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)
    handler.setLevel(level)

    return handler


def get_all_handlers(logger: logging.Logger) -> List[logging.Handler]:
    handlers: List[logging.Handler] = []
    if logger.handlers:
        handlers.extend(list(logger.handlers))
    while logger.parent:
        logger = logger.parent
        if logger.handlers:
            handlers.extend(list(logger.handlers))
    return handlers


def check_formatter_equality(
    this: Optional[logging.Formatter], other: Optional[logging.Formatter]
) -> bool:
    if this is None and other is None:
        return True
    if this is None or other is None:
        return False
    else:
        return this._fmt == other._fmt and this.datefmt == other.datefmt


def check_handler_equality(this: logging.Handler, other: logging.Handler) -> bool:
    if type(this) is not type(other):
        return False
    if this.level != other.level:
        return False
    if this.get_name() != other.get_name():
        return False
    if not check_formatter_equality(this.formatter, other.formatter):
        return False
    return True


def enable_stdout_logging(
    logger: StrOrLogger, format: StrOrFormatter = DEFAULT_LOGGING_FORMAT, level: LogLevel = "INFO"
) -> logging.Logger:
    if not isinstance(logger, logging.Logger):
        logger = logging.getLogger(logger)

    handler = get_stdout_handler(format=format, level=level)

    if any([check_handler_equality(handler, existing) for existing in get_all_handlers(logger)]):
        return logger

    logger.addHandler(handler)
    return logger


class LoggingMixin(object):
    @cached_property
    def logger(self) -> logging.Logger:
        name = f"{self.__class__.__module__}.{self.__class__.__name__}"
        return get_logger(name)

    @property
    def log(self) -> logging.Logger:
        return self.logger

    def log_stacktrace(self, message: str, error: BaseException):
        self.logger.error(message)
        self.logger.exception(error)

    def enable_stdout_logging(self):
        enable_stdout_logging(self.logger)



__all__ = [
    "JSONContent",
    "JSONReference",
    "PutJSONToFileRequest",
    "PutJSONToFileResponse",
    "GetJSONFromFileRequest",
    "GetJSONFromFileResponse",
    "DataSyncTask",
    "DataSyncConfig",
    "DataSyncRequest",
    "DataSyncResponse",
    "BatchDataSyncRequest",
    "BatchDataSyncResponse",
    "PrepareBatchDataSyncRequest",
    "PrepareBatchDataSyncResponse",
]

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import marshmallow as mm

from aibs_informatics_core.models.aws.efs import EFSPath
from aibs_informatics_core.models.aws.s3 import S3KeyPrefix, S3Path
from aibs_informatics_core.models.base import (
    BooleanField,
    CustomStringField,
    IntegerField,
    ListField,
    PathField,
    RawField,
    SchemaModel,
    UnionField,
    custom_field,
)
from aibs_informatics_core.utils.json import JSON


@dataclass
class JSONContent(SchemaModel):
    content: JSON = custom_field(mm_field=RawField())


@dataclass
class JSONReference(SchemaModel):
    path: Union[S3Path, Path] = custom_field(
        mm_field=UnionField([(S3Path, S3Path.as_mm_field()), (Path, PathField())])
    )


@dataclass
class PutJSONToFileRequest(JSONContent):
    path: Optional[Union[S3Path, Path]] = custom_field(
        default=None, mm_field=UnionField([(S3Path, S3Path.as_mm_field()), (Path, PathField())])
    )


@dataclass
class PutJSONToFileResponse(JSONReference):
    pass


@dataclass
class GetJSONFromFileRequest(JSONReference):
    pass


@dataclass
class GetJSONFromFileResponse(JSONContent):
    pass


@dataclass
class DataSyncTask(SchemaModel):
    source_path: Union[S3Path, EFSPath, Path] = custom_field(
        mm_field=UnionField(
            [
                (S3Path, S3Path.as_mm_field()),
                (EFSPath, EFSPath.as_mm_field()),
                ((Path, str), PathField()),
            ]
        )
    )
    destination_path: Union[S3Path, EFSPath, Path] = custom_field(
        mm_field=UnionField(
            [
                (S3Path, S3Path.as_mm_field()),
                (EFSPath, EFSPath.as_mm_field()),
                ((Path, str), PathField()),
            ]
        )
    )
    source_path_prefix: Optional[S3KeyPrefix] = custom_field(
        default=None, mm_field=CustomStringField(S3KeyPrefix)
    )


@dataclass
class RemoteToLocalConfig(SchemaModel):
    # Use a custom intermediate tmp dir when syncing an s3 object to a local filesystem
    # instead of using boto3's implementation which creates a part file (e.g. *.6eF5b5da)
    # in SAME parent dir as the desired destination path.
    use_custom_tmp_dir: bool = custom_field(default=False, mm_field=BooleanField())
    custom_tmp_dir: Optional[Union[EFSPath, Path]] = custom_field(
        default=None,
        mm_field=UnionField(
            [
                (EFSPath, EFSPath.as_mm_field()),
                ((Path, str), PathField()),
            ]
        ),
    )


@dataclass
class DataSyncConfig(SchemaModel):
    max_concurrency: int = custom_field(default=25, mm_field=IntegerField())
    retain_source_data: bool = custom_field(default=True, mm_field=BooleanField())
    require_lock: bool = custom_field(default=False, mm_field=BooleanField())
    force: bool = custom_field(default=False, mm_field=BooleanField())
    size_only: bool = custom_field(default=False, mm_field=BooleanField())
    fail_if_missing: bool = custom_field(default=True, mm_field=BooleanField())
    include_detailed_response: bool = custom_field(default=False, mm_field=BooleanField())
    remote_to_local_config: RemoteToLocalConfig = custom_field(
        default_factory=RemoteToLocalConfig,
        mm_field=RemoteToLocalConfig.as_mm_field(),
    )


@dataclass
class DataSyncRequest(DataSyncConfig, DataSyncTask):  # type: ignore[misc]
    @property
    def config(self) -> DataSyncConfig:
        return DataSyncConfig(
            max_concurrency=self.max_concurrency,
            retain_source_data=self.retain_source_data,
            require_lock=self.require_lock,
            force=self.force,
            size_only=self.size_only,
            fail_if_missing=self.fail_if_missing,
            include_detailed_response=self.include_detailed_response,
            remote_to_local_config=self.remote_to_local_config,
        )

    @property
    def task(self) -> DataSyncTask:
        return DataSyncTask(
            source_path=self.source_path,
            destination_path=self.destination_path,
            source_path_prefix=self.source_path_prefix,
        )


@dataclass
class DataSyncResult(SchemaModel):
    bytes_transferred: int = 0
    files_transferred: int = 0

    def add_bytes_transferred(self, bytes_transferred: int) -> None:
        self.bytes_transferred += bytes_transferred

    def add_files_transferred(self, files_transferred: int) -> None:
        self.files_transferred += files_transferred


@dataclass
class DataSyncResponse(SchemaModel):
    request: DataSyncRequest = custom_field(mm_field=DataSyncRequest.as_mm_field())
    result: DataSyncResult = custom_field(mm_field=DataSyncResult.as_mm_field())


@dataclass
class BatchDataSyncRequest(SchemaModel):
    requests: Union[List[DataSyncRequest], S3Path] = custom_field(
        mm_field=UnionField(
            [
                (list, ListField(DataSyncRequest.as_mm_field())),
                (S3Path, S3Path.as_mm_field()),
            ]
        )
    )
    allow_partial_failure: bool = custom_field(default=False, mm_field=BooleanField())

    @classmethod
    @mm.pre_load
    def _handle_single_flattened_request(cls, data: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        if DataSyncRequest.is_valid(data=data, many=False, partial=False):
            data = {
                "requests": [data],
                "allow_partial_failure": False,
                "include_detailed_response": False,
            }
        return data


@dataclass
class BatchDataSyncResult(DataSyncResult):
    total_requests_count: int = 0
    successful_requests_count: int = 0
    failed_requests_count: int = 0

    def increment_successful_requests_count(self, increment: int = 1) -> None:
        self.successful_requests_count += increment
        self.total_requests_count += increment

    def increment_failed_requests_count(self, increment: int = 1) -> None:
        self.failed_requests_count += increment
        self.total_requests_count += increment


@dataclass
class BatchDataSyncResponse(SchemaModel):
    result: BatchDataSyncResult = custom_field(mm_field=BatchDataSyncResult.as_mm_field())
    failed_requests: Optional[List[DataSyncRequest]] = custom_field(default=None)

    def add_failed_request(self, request: DataSyncRequest) -> None:
        if self.failed_requests is None:
            self.failed_requests = []
        self.failed_requests.append(request)


@dataclass
class PrepareBatchDataSyncRequest(DataSyncRequest):
    batch_size_bytes_limit: Optional[int] = custom_field(default=None, mm_field=IntegerField())
    temporary_request_payload_path: Optional[S3Path] = custom_field(
        default=None, mm_field=S3Path.as_mm_field()
    )


@dataclass
class PrepareBatchDataSyncResponse(SchemaModel):
    requests: List[BatchDataSyncRequest] = custom_field(
        mm_field=ListField(BatchDataSyncRequest.as_mm_field())
    )



from __future__ import annotations

import json
import logging
import os
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generic, Optional, Type, TypeVar, Union

from aibs_informatics_core.collections import PostInitMixin
from aibs_informatics_core.env import EnvBaseMixins
from aibs_informatics_core.models.aws.s3 import S3Path
from aibs_informatics_core.models.base import ModelProtocol
from aibs_informatics_core.utils.json import JSON, JSONObject, is_json_str, load_json_object

logger = logging.getLogger(__name__)


REQUEST = TypeVar("REQUEST", bound=ModelProtocol)
RESPONSE = TypeVar("RESPONSE", bound=ModelProtocol)

E = TypeVar("E", bound="BaseExecutor")


@dataclass  # type: ignore[misc] # mypy #5374
class BaseExecutor(EnvBaseMixins, PostInitMixin, Generic[REQUEST, RESPONSE]):
    @abstractmethod
    def handle(self, request: REQUEST) -> Optional[RESPONSE]:  # pragma: no cover
        """Core logic for handling request

        NOT IMPLEMENTED

        Args:
            request (API_REQUEST): Request object expected

        Returns:
            Optional[API_RESPONSE]: response object returned, Optional
        """
        raise NotImplementedError("Must implement handler logic here")

    # --------------------------------------------------------------------
    # Request & Response De-/Serialization
    # --------------------------------------------------------------------

    @classmethod
    def get_request_cls(cls) -> Type[REQUEST]:
        return cls.__orig_bases__[0].__args__[0]  # type: ignore

    @classmethod
    def get_response_cls(cls) -> Type[RESPONSE]:
        return cls.__orig_bases__[0].__args__[1]  # type: ignore

    @classmethod
    def deserialize_request(cls, request: JSON) -> REQUEST:
        """Deserialize a raw request

        Supported request types:
        Union[JSONObject, S3Path, str, Path, REQUEST]
            1. dict object
            2. S3 Path object (must implement s3 deserialization method)
            3. stringified object
            4. path to file of json object

        Args:
            request (Union[JSONObject, S3Path, str, Path]): raw request

        Returns:
            REQUEST: Deserialized request
        """
        if isinstance(request, cls.get_request_cls()):
            return request

        request = cls.load_input(request)
        if isinstance(request, dict):
            try:
                return cls.get_request_cls().from_dict(request)
            except Exception as e:
                logger.error(f"Could not deserialize object. error: {e}")
                raise e
        else:
            raise ValueError(f"Cannot deserialize request: {request}, type: {type(request)}")

    @classmethod
    def serialize_response(cls, response: RESPONSE) -> JSONObject:
        """Serializes response object into JSON

        Args:
            response (API_RESPONSE): The returned response object

        Returns:
            JSON: serialized form of response object
        """
        return response.to_dict()

    @classmethod
    def load_input(cls, input: Any) -> JSON:
        """Loads input from various sources

        Supported input types:
            1. dict object
            2. S3 Path object (must implement s3 deserialization method)
            3. stringified object
            4. path to file of json object

        Args:
            input (Any): raw input. Can be any of the supported types.

        Returns:
            JSON: content loaded from input
        """
        if isinstance(input, dict):
            return input
        elif isinstance(input, str) and is_json_str(input):
            return load_json_object(input)
        elif isinstance(input, S3Path) or (isinstance(input, str) and S3Path.is_valid(input)):
            return cls.load_input__remote(S3Path(input))
        elif isinstance(input, (str, Path)):
            return cls.load_input__file(Path(input))
        else:
            raise ValueError(f"Cannot read input: {input}, type: {type(input)}")

    @classmethod
    def load_input__remote(cls, remote_path: S3Path) -> JSON:
        """Reads input from S3 location

        Args:
            s3_path (S3Path): S3 location to read input from

        Returns:
            JSON: JSON data read from S3
        """
        raise NotImplementedError(
            f"{cls.__name__} has not provided an implementation for this form of input"
        )

    @classmethod
    def load_input__file(cls, local_path: Path) -> JSON:
        """Reads input from file

        Args:
            local_path (Path): Local path to read input from

        Returns:
            JSON: JSON data read from file
        """
        if not local_path.exists():
            raise ValueError(f"local path specified {local_path} does not exist")
        if local_path.is_dir():
            raise ValueError(f"local path specified {local_path} cannot be read. Must be a file.")
        return load_json_object(local_path)

    @classmethod
    def write_output(cls, output: JSON, path: Union[str, Path]) -> None:
        """Writes output to location

        Args:
            output (JSON): serialized output object
            path (Path | str): location to write output to. Can be S3 or local file
        """

        if isinstance(path, S3Path) or (isinstance(path, str) and S3Path.is_valid(path)):
            return cls.write_output__remote(output, S3Path(path))
        else:
            # Here we assume that it could be string or path to file
            return cls.write_output__file(output, Path(path))

    @classmethod
    def write_output__remote(cls, output: JSON, remote_path: S3Path) -> None:
        """Writes output to Remote location

        Args:
            output (RESPONSE): output object
            remote_path (S3Path): S3 location to write response to
        """
        raise NotImplementedError(
            f"{cls.__name__} has not provided an implementation for this form of serialization"
        )

    @classmethod
    def write_output__file(cls, output: JSON, local_path: Path) -> None:
        """Writes output to file

        Args:
            output (JSON): serialized output json
            local_path (S3Path): Local path to write response to
        """

        if local_path.is_dir() or (
            local_path.parent.exists() and not os.access(local_path.parent, os.W_OK)
        ):
            raise ValueError(
                f"local path specified {local_path} cannot be written to. Must be a file "
            )
        local_path.parent.mkdir(parents=True, exist_ok=True)
        local_path.write_text(json.dumps(output, sort_keys=True))

    @classmethod
    def get_executor_name(cls) -> str:
        """Returns a distinguishable name of the executor class

        By default, it returns class name.

        """
        return cls.__name__

    @classmethod
    def build_from_env(cls: Type[E], **kwargs) -> E:
        """Creates an executor from environment

        Must be able to create an executor from environment variables

        Returns:
            E: Executor object
        """
        return cls(**kwargs)

    @classmethod
    def run_executor(
        cls, input: JSON, output_location: Optional[Union[str, Path]] = None, **kwargs
    ) -> Optional[JSON]:
        """Runs an executor

        Args:
            input (JSON): input to executor
            output_location (Optional[str], optional): Optional output location to write
                response to. Defaults to None.
        """
        executor = cls.build_from_env(**kwargs)

        request = cls.deserialize_request(input)
        response = executor.handle(request)
        output = executor.serialize_response(response) if response else None

        if output_location:
            if output is None:
                logger.warning(
                    "Response is None but output location is specified. Writing empty dict"
                )
            executor.write_output(output or {}, output_location)

        return output



from aibs_informatics_core.collections import StrEnum


class Status(StrEnum):
    AWAITING_TRIGGER = "AWAITING_TRIGGER"
    SUBMITTED = "SUBMITTED"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    ABORTED = "ABORTED"

    def is_complete(self) -> bool:
        return self == Status.COMPLETED or self == Status.FAILED or self == Status.ABORTED



__all__ = [
    "StringifiedResolvable",
    "StringifiedDownloadable",
    "StringifiedUploadable",
    "ResolvableBase",
    "ResolvableAction",
    "Resolvable",
    "S3Resolvable",
    "Uploadable",
    "get_resolvable_from_value",
]

import re
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, ClassVar, Dict, Generic, Optional, Pattern, Sequence, Type, TypeVar, Union

import marshmallow as mm

from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.exceptions import ValidationError
from aibs_informatics_core.models.aws.s3 import S3URI
from aibs_informatics_core.models.base import SchemaModel, custom_field
from aibs_informatics_core.utils.hashing import sha256_hexdigest

ACTION_PATTERN = " @ "
URI_PATTERN = r"(?:\w+:)?(?:\/?\/?)[^\s]+"

PATH_PATTERN = r"(?:\/)?(?:[^\0]+\/?)*"
REMOTE_PATTERN = r"[^\s]*"
LOCAL_PATTERN = r"(?:\/)?(?:[^/\0]+\/?)*"

T = TypeVar("T", bound=str)
STRINGIFIED_RESOLVABLE = TypeVar("STRINGIFIED_RESOLVABLE", bound="StringifiedResolvable")
RESOLVABLE = TypeVar("RESOLVABLE", bound="ResolvableBase")


class StringifiedResolvable(ValidatedStr):
    """Stringified representation of a Resolvable object.

    Resolvables have two components:
        source:
        destination:

    Examples:
        "s3://bucket/key"               ->    source:  "s3://bucket/key"
                                              destination: "tmp{SHA256HEX}"
        "s3://bucket/key@version"       ->    source:  "s3://bucket/key@version"
                                              destination: "tmp{SHA256HEX}"
        "s3://bucket/key @ /tmp/xxx"    ->    source:  "s3://bucket/key"
                                              destination: "/tmp/xxx"
        "@ /tmp/xxx @ s3://bucket/key"    ->    source: "/tmp/xxx"
                                              destination:  "s3://bucket/key"
    """

    regex_pattern: ClassVar[Pattern] = re.compile(
        rf"(?:({URI_PATTERN})(?:{ACTION_PATTERN})({URI_PATTERN}))|({URI_PATTERN})"
    )

    @property
    def source(self) -> str:
        return self.get_match_groups()[0] or self.get_match_groups()[-1]

    @property
    def destination(self) -> Optional[str]:
        return self.get_match_groups()[1]

    @property
    @abstractmethod
    def local(self) -> Optional[str]:
        raise NotImplementedError("please implement")

    @property
    @abstractmethod
    def remote(self) -> Optional[str]:
        raise NotImplementedError("please implement")

    @classmethod
    def from_components(
        cls: Type[STRINGIFIED_RESOLVABLE], source: str, destination: Optional[str]
    ) -> STRINGIFIED_RESOLVABLE:
        if destination:
            return cls(f"{source}{ACTION_PATTERN}{destination}")
        return cls(source)


class StringifiedDownloadable(StringifiedResolvable):
    @property
    def local(self) -> str:
        return self.destination or f"tmp{sha256_hexdigest(self.remote)[:8]}"

    @property
    def remote(self) -> str:
        return self.source


class StringifiedUploadable(StringifiedResolvable):
    @property
    def local(self) -> str:
        return self.source

    @property
    def remote(self) -> Optional[str]:
        return self.destination


class ResolvableAction(str, Enum):
    LOCALIZE = "LOCALIZE"
    DELOCALIZE = "DELOCALIZE"


# TODO: rename to rename resolvable subclasses to uploadable
#       Also add action and maybe a type field (s3, gfs, etc.)
@dataclass  # type: ignore[misc] # mypy #5374
class ResolvableBase(SchemaModel, Generic[T]):
    local: str = custom_field()
    remote: T = custom_field()

    @classmethod
    def get_action(cls) -> ResolvableAction:
        return ResolvableAction.LOCALIZE

    @classmethod
    def get_resolvable_type(cls: Type[RESOLVABLE]) -> Type[T]:
        return cls.__orig_bases__[0].__args__[0]  # type: ignore[attr-defined]

    @classmethod
    def from_any(
        cls: Type[RESOLVABLE],
        value: Any,
        default_local: Optional[str] = None,
        default_remote: Optional[T] = None,
    ) -> RESOLVABLE:
        if isinstance(value, cls):
            return value
        elif isinstance(value, dict):
            obj = cls.from_dict(value, partial=True)
            if obj.local is None or cls.is_missing(obj.local):
                if default_local is None:
                    raise ValueError(f"Local is None for {value}. No default provided")
                obj.local = default_local
            obj.remote = obj.remote or default_remote
            obj.to_dict(validate=True)
            return obj
        elif isinstance(value, str):
            return cls.from_str(value, default_local=default_local, default_remote=default_remote)
        else:
            raise ValueError(f"Value {value} is not a dict or str. Cannot create {cls}")

    @classmethod
    def from_str(
        cls: Type[RESOLVABLE],
        value: str,
        default_local: Optional[str] = None,
        default_remote: Optional[T] = None,
    ) -> RESOLVABLE:
        str_resolvable_cls = (
            StringifiedUploadable
            if cls.get_action() == ResolvableAction.DELOCALIZE
            else StringifiedDownloadable
        )

        stringified_resolvable = str_resolvable_cls(value)

        local = stringified_resolvable.local or default_local
        assert local is not None, f"Local is None for {value}"
        if stringified_resolvable.remote:
            remote = cls.get_resolvable_type()(stringified_resolvable.remote)
        else:
            remote = default_remote
        return cls(local=local, remote=remote)

    def to_str(self) -> Union[StringifiedDownloadable, StringifiedUploadable]:
        if self.get_action() == ResolvableAction.LOCALIZE:
            return StringifiedDownloadable.from_components(
                source=self.remote, destination=self.local
            )
        else:
            return StringifiedUploadable.from_components(
                source=self.local, destination=self.remote
            )

    @classmethod
    @mm.post_dump
    def _post_dump__inject_action(cls, data: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        data["action"] = cls.get_action()
        return data

    @classmethod
    @mm.pre_load
    def _pre_load__pop_action(cls, data: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        if "action" in data:
            assert data["action"] == cls.get_action()
            del data["action"]
        return data


R = TypeVar("R", bound=ResolvableBase)


def get_resolvable_from_value(value: Any, resolvable_classes: Sequence[Type[R]]) -> R:
    """Construct resolvable object from string or dict

    If a resolvable object is provided, it returned immediately.

    Args:
        value (Any): string or dictionary. Raises error otherwise
        resolvable_classes (Sequence[Type[R]]): sequence of resolvable classes.
            They are evaluated in the order they are provided.

    Raises:
        ValueError: The input is not a valid type.
        mm.ValidationError: No resolvable object could be constructed from input

    Returns:
        R: resolvable object
    """
    if any([isinstance(value, _) for _ in resolvable_classes]):
        return value

    if not isinstance(value, (str, dict)):
        raise ValueError(
            f"Value {value} is not a dict or str. Cannot create any of these: {resolvable_classes}"
        )

    errors: Dict[str, Exception] = {}
    for resolvable_class in resolvable_classes:
        try:
            return resolvable_class.from_any(value)
        except (mm.ValidationError, ValidationError, ValueError) as e:
            errors[resolvable_class.__name__] = e
    else:
        raise mm.ValidationError(
            {**{"ALL": f"Could not create any {resolvable_classes} from {value}"}, **errors}, "n/a"
        )


@dataclass
class Resolvable(ResolvableBase[str]):
    remote: str = custom_field()


@dataclass
class S3Resolvable(ResolvableBase[S3URI]):
    remote: S3URI = custom_field(mm_field=S3URI.as_mm_field())


@dataclass
class Uploadable(Resolvable):
    @classmethod
    def get_action(cls) -> ResolvableAction:
        return ResolvableAction.DELOCALIZE



__all__ = ["HTTPParameters"]


import ast
import json
import urllib.parse
from base64 import urlsafe_b64decode, urlsafe_b64encode
from dataclasses import dataclass
from typing import Dict, Optional

from aibs_informatics_core.exceptions import ValidationError
from aibs_informatics_core.utils.json import JSON
from aibs_informatics_core.utils.logging import get_logger

logger = get_logger()


QUERY_PARAMS_KEY = "data"


@dataclass
class HTTPParameters:
    route_params: Optional[Dict[str, JSON]]
    query_params: Optional[Dict[str, JSON]]
    request_body: Optional[Dict[str, JSON]]

    @property
    def stringified_route_params(self) -> Dict[str, str]:
        return self.to_stringified_route_params(self.route_params)

    @property
    def stringified_query_params(self) -> Dict[str, str]:
        return self.to_stringified_query_params(self.query_params)

    @property
    def stringified_request_body(self) -> Optional[str]:
        return self.to_stringified_request_body(self.request_body)

    @property
    def merged_params(self) -> Dict[str, JSON]:
        request_json = {}
        # First add route parameters
        if self.route_params:
            request_json.update(self.route_params)
        # Next add query parameters
        if self.query_params:
            request_json.update(self.query_params)
        # finally add request body
        if self.request_body:
            request_json.update(self.request_body)
        return request_json

    @classmethod
    def from_stringified_route_params(
        cls, parameters: Optional[Dict[str, str]]
    ) -> Dict[str, JSON]:
        evaluated_params = dict()

        input_params = parameters or dict()
        # literal_eval as much as we can to python primitives, our schema will take care of rest.
        for k, v in input_params.items():
            try:
                evaluated_params[k] = ast.literal_eval(urllib.parse.unquote(v))
            except Exception:
                logger.warning(f"could not parse {v}. setting as is.")
                evaluated_params[k] = v
        return evaluated_params

    @classmethod
    def to_stringified_route_params(cls, parameters: Optional[Dict[str, JSON]]) -> Dict[str, str]:
        string_params = {}
        for k, v in (parameters or {}).items():
            try:
                string_params[k] = str(v)
            except Exception as e:
                raise ValidationError(f"Couldn't stringify {k} value=?") from e
        return string_params

    @classmethod
    def from_stringified_query_params(
        cls, parameters: Optional[Dict[str, str]]
    ) -> Dict[str, JSON]:
        if parameters is None or len(parameters) == 0:
            return {}
        elif QUERY_PARAMS_KEY not in parameters:
            raise ValidationError(
                f"Stringified Query parameters MUST have {QUERY_PARAMS_KEY} field"
            )
        parameters_str = urlsafe_b64decode(parameters[QUERY_PARAMS_KEY].encode()).decode()
        return json.loads(parameters_str)

    @classmethod
    def to_stringified_query_params(cls, parameters: Optional[Dict[str, JSON]]) -> Dict[str, str]:
        if parameters is None or len(parameters) == 0:
            return {}
        parameters_str = json.dumps(parameters, sort_keys=True)
        return {QUERY_PARAMS_KEY: urlsafe_b64encode(parameters_str.encode()).decode()}

    @classmethod
    def to_stringified_request_body(cls, parameters: Optional[Dict[str, JSON]]) -> Optional[str]:
        if parameters is None or len(parameters) == 0:
            return None
        return json.dumps(parameters, sort_keys=True)

    @classmethod
    def from_stringified_request_body(cls, parameters: Optional[str]) -> Dict[str, JSON]:
        if parameters is None or len(parameters) == 0:
            return {}
        return json.loads(parameters)

    @classmethod
    def from_http_request(
        cls,
        stringified_route_params: Optional[Dict[str, str]],
        stringified_query_params: Optional[Dict[str, str]],
        stringified_request_body: Optional[str],
    ) -> "HTTPParameters":
        return HTTPParameters(
            route_params=cls.from_stringified_route_params(stringified_route_params),
            query_params=cls.from_stringified_query_params(stringified_query_params),
            request_body=cls.from_stringified_request_body(stringified_request_body),
        )



__all__ = [
    "expandvars",
    "find_all_paths",
    "get_env_var",
    "set_env_var",
    "to_env_var_dict",
    "to_env_var_list",
    "order_env_vars",
    "generate_env_file_content",
    "write_env_file",
    "EnvVarDictItemUpper",
    "EnvVarDictItemLower",
    "EnvVarTupleItem",
    "EnvVarItem",
    "EnvVarItemType",
    "EnvVarFormat",
    "EnvVarCollection",
]

import os
import re
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypedDict,
    Union,
    cast,
    overload,
)


def expandvars(path, default=None, skip_escaped=False):
    """Expand environment variables of form $var and ${var}.
    If parameter 'skip_escaped' is True, all escaped variable references
    (i.e. preceded by backslashes) are skipped.
    Unknown variables are set to 'default'. If 'default' is None,
    they are left unchanged.
    """

    def replace_var(m):
        return os.environ.get(m.group(2) or m.group(1), m.group(0) if default is None else default)

    env_var_pattern = re.compile((r"(?<!\\)" if skip_escaped else "") + r"\$(\w+|\{([^}]*)\})")
    return re.sub(env_var_pattern, replace_var, path)


def find_all_paths(
    root: Union[str, Path], include_dirs: bool = True, include_files: bool = True
) -> List[str]:
    """Find all paths below root path

    Args:
        root (str | Path): root path to start
        include_dirs (bool, optional): Whether to include directories. Defaults to True.
        include_files (bool, optional): whether to include files. Defaults to True.
    Returns:
        List[str]: list of paths found under root
    """
    paths = []
    str_root = str(root) if isinstance(root, Path) else root
    if os.path.isfile(str_root) and include_files:
        paths.append(str_root)
    for parent, dirs, files in os.walk(str_root):
        if include_dirs:
            paths.extend([os.path.join(parent, name) for name in dirs])
        if include_files:
            paths.extend([os.path.join(parent, name) for name in files])
    return paths


@overload
def get_env_var(*keys: str) -> Optional[str]: ...  # pragma: no cover


@overload
def get_env_var(*keys: str, default_value: Literal[None]) -> Optional[str]: ...  # pragma: no cover


@overload
def get_env_var(*keys: str, default_value: str) -> str: ...  # pragma: no cover


def get_env_var(*keys: str, default_value: Optional[str] = None) -> Optional[str]:
    """get env variable using one of keys (sorted by priority)

    Arguments:
        keys (Tuple[str]): list of env keys to check
            (sorted based on fallback priority)
        default_value: value to use if none are found. Defaults to none

    Returns:
        Optional[str]:
    """
    for key in keys:
        val = os.environ.get(key)
        if val:
            return val
    else:
        return default_value


def set_env_var(key: str, value: str):
    os.environ[key] = value


class EnvVarDictItemUpper(TypedDict):
    Key: str
    Value: str


class EnvVarDictItemLower(TypedDict):
    key: str
    value: str


EnvVarTupleItem = Tuple[str, str]


@dataclass
class EnvVarItem:
    key: str
    value: str

    @overload
    def to_dict(self, lower: Literal[False] = False) -> EnvVarDictItemUpper: ...

    @overload
    def to_dict(self, lower: Literal[True]) -> EnvVarDictItemLower: ...

    def to_dict(self, lower: bool = False) -> Union[EnvVarDictItemUpper, EnvVarDictItemLower]:
        if lower:
            return EnvVarDictItemLower(key=self.key, value=self.value)
        else:
            return EnvVarDictItemUpper(Key=self.key, Value=self.value)

    def to_tuple(self) -> EnvVarTupleItem:
        return (self.key, self.value)

    @classmethod
    def from_any(cls, value: Any) -> "EnvVarItem":
        if isinstance(value, cls):
            return value
        elif isinstance(value, tuple):
            assert len(value) == 2
            return cls(*value)
        elif isinstance(value, dict):
            return cls(**{k.lower(): v for k, v in value.items()})
        else:
            raise ValueError(f"Invalid env_var type: {type(value)}")


EnvVarItemType = Union[
    EnvVarItem,
    EnvVarTupleItem,
    EnvVarDictItemUpper,
    EnvVarDictItemLower,
]


class EnvVarFormat(Enum):
    OBJECT = "object"
    TUPLE = "tuple"
    DICT_LOWER = "dict_lower"
    DICT_UPPER = "dict_upper"


EnvVarSequence = Union[
    Sequence[EnvVarItem],
    Sequence[EnvVarTupleItem],
    Sequence[EnvVarDictItemUpper],
    Sequence[EnvVarDictItemLower],
]

EnvVarCollection = Union[
    EnvVarSequence,
    Dict[str, str],
]


def to_env_var_dict(env_vars: EnvVarCollection) -> Dict[str, str]:
    """Converts env vars to a dict

    Args:
        env_vars (Tuple[str, str] | Dict[str, str]): env vars to convert

    Returns:
        Dict[str, str]: dict of env vars
    """
    if isinstance(env_vars, dict):
        return env_vars

    env_var_dict = {}
    env_var_items = [EnvVarItem.from_any(env_var) for env_var in env_vars]
    for env_var_item in env_var_items:
        env_var_dict[env_var_item.key] = env_var_item.value
    return env_var_dict


@overload
def to_env_var_list(env_vars: EnvVarCollection) -> List[EnvVarTupleItem]: ...


@overload
def to_env_var_list(
    env_vars: EnvVarCollection, env_var_format: Literal[EnvVarFormat.TUPLE]
) -> List[EnvVarTupleItem]: ...


@overload
def to_env_var_list(
    env_vars: EnvVarCollection, env_var_format: Literal[EnvVarFormat.OBJECT]
) -> List[EnvVarItem]: ...


@overload
def to_env_var_list(
    env_vars: EnvVarCollection, env_var_format: Literal[EnvVarFormat.DICT_LOWER]
) -> List[EnvVarDictItemLower]: ...


@overload
def to_env_var_list(
    env_vars: EnvVarCollection, env_var_format: Literal[EnvVarFormat.DICT_UPPER]
) -> List[EnvVarDictItemUpper]: ...


def to_env_var_list(
    env_vars: EnvVarCollection, env_var_format: EnvVarFormat = EnvVarFormat.TUPLE
) -> EnvVarSequence:
    if isinstance(env_vars, dict):
        env_var_list = [EnvVarItem(key, value) for key, value in env_vars.items()]
    else:
        env_var_list = [EnvVarItem.from_any(env_var) for env_var in env_vars]

    if env_var_format == EnvVarFormat.OBJECT:
        return env_var_list
    elif env_var_format == EnvVarFormat.TUPLE:
        env_var_tuple_list = [_.to_tuple() for _ in env_var_list]
        return env_var_tuple_list
    elif env_var_format == EnvVarFormat.DICT_LOWER:
        env_var_dict_list = [_.to_dict(lower=True) for _ in env_var_list]
        return env_var_dict_list
    elif env_var_format == EnvVarFormat.DICT_UPPER:
        env_var_dict_list = [_.to_dict(lower=False) for _ in env_var_list]
        return env_var_dict_list
    else:
        raise ValueError(f"Invalid env_var_format: {env_var_format}")


@overload
def order_env_vars(
    env_vars: EnvVarCollection, env_var_format: Literal[EnvVarFormat.OBJECT]
) -> List[EnvVarItem]: ...


@overload
def order_env_vars(
    env_vars: EnvVarCollection, env_var_format: Literal[EnvVarFormat.TUPLE]
) -> List[EnvVarTupleItem]: ...


@overload
def order_env_vars(
    env_vars: EnvVarCollection, env_var_format: Literal[EnvVarFormat.DICT_LOWER]
) -> List[EnvVarDictItemLower]: ...


@overload
def order_env_vars(
    env_vars: EnvVarCollection, env_var_format: Literal[EnvVarFormat.DICT_UPPER]
) -> List[EnvVarDictItemUpper]: ...


def order_env_vars(
    env_vars: EnvVarCollection, env_var_format: EnvVarFormat = EnvVarFormat.TUPLE
) -> EnvVarCollection:
    """Resolve the order of environment variables based on their dependencies.

    This function performs a topological sort on the environment variables,
    ordering them based on the dependencies found in their values. If a circular
    dependency is detected, a ValueError will be raised.

    Arguments:
        environment (Dict[str, str]): A dictionary where keys are environment variable names
            and values are their corresponding values. Dependencies are represented as
            ${var_name} or $var_name in the values.

    Returns:
        List[Tuple[str, str]]: A list of tuples, where each tuple is a pair of
            environment variable name and its value, ordered based on their dependencies.

    Raises:
        ValueError: If a circular dependency is detected among the environment variables.
    """
    # Create a list of environment keys
    env_var_dict = to_env_var_dict(env_vars)
    env_keys = list(env_var_dict.keys())

    # Initialize dependency maps
    env_key_deps: Dict[str, Set[str]] = {k: set() for k in env_keys}
    env_key_deps_rev: Dict[str, Set[str]] = {k: set() for k in env_keys}

    # Populate dependency maps
    for k, v in env_var_dict.items():
        for match in cast(List[Tuple[str, str]], re.findall(r"(?:\$\{([^\}]+)\}|\$([\w]+))", v)):
            if (value := match[0] or match[1]) in env_keys:
                # Add dependencies
                env_key_deps[k].add(value)
                env_key_deps_rev[value].add(k)

    # Sort environment keys based on the number of dependencies
    unordered_env_keys = sorted(env_keys, key=lambda k: len(env_key_deps[k]), reverse=False)

    # Initialize the list of ordered environment key-value pairs
    ordered_env_key_pairs: List[EnvVarItem] = []

    # Order environment keys
    while unordered_env_keys:
        for i, k in enumerate(unordered_env_keys):
            # If a key has no dependencies, add it to the ordered list
            if not env_key_deps[k]:
                ordered_env_key_pairs.append(EnvVarItem(k, env_var_dict[k]))
                unordered_env_keys.pop(i)

                # Update dependency maps
                for rev_k in env_key_deps_rev.pop(k, set()):
                    env_key_deps[rev_k].discard(k)
                break
        else:
            # If a circular dependency is detected, raise an error
            raise ValueError("Circular dependency detected in environment variables")
    return to_env_var_list(ordered_env_key_pairs, env_var_format=env_var_format)  # type: ignore[call-overload]  # dynamic env var format doesnt work with mypy


def generate_env_file_content(env_vars: EnvVarCollection) -> str:
    """Generate the content of an environment file.

    Example:
        Given environment variables:
            - X = "${Y}/A"
            - Y = "${Z}/B"
            - Z = "C"
        Output:
            export Z=C
            export Y=${Z}/B
            export X=${Y}/A

    Args:
        env_vars (EnvVarCollection): A collection of environment variables.

    Returns:
        str: The content of the environment file.
    """
    return "\n".join(
        [
            f'export {k}="{v}"'
            # sort the remaining environment variables to proper order
            for k, v in order_env_vars(env_vars, env_var_format=EnvVarFormat.TUPLE)
        ]
    )


def write_env_file(env_vars: EnvVarCollection, path: Union[str, Path]):
    """Write environment variables to a file.

    Args:
        env_vars (EnvVarCollection): A collection of environment variables.
        path (Union[str, Path]): The path of the environment file.
    """
    path = Path(path)
    path.write_text(generate_env_file_content(env_vars))


@contextmanager
def env_var_overrides(*env_vars: EnvVarItemType):
    """Temporarily override environment variables"""
    original_env_vars = {}
    env_var_dict = to_env_var_dict(env_vars)  # type: ignore[arg-type]
    for key, value in env_var_dict.items():
        original_env_vars[key] = os.environ.get(key)
        os.environ[key] = value
    try:
        yield
    finally:
        for k, v in original_env_vars.items():
            if v is None:
                del os.environ[k]
            else:
                os.environ[k] = v



__all__ = [
    "camelcase",
    "is_prefixed",
    "is_suffixed",
    "lowercase",
    "pascalcase",
    "removeprefix",
    "removesuffix",
    "snakecase",
    "spinalcase",
    "uppercase",
]


from dataclasses_json import stringcase


def is_prefixed(value: str, prefix: str) -> bool:
    return value[: len(prefix)] == prefix


def is_suffixed(value: str, suffix: str) -> bool:
    return value[-len(suffix) :] == suffix


def removeprefix(value: str, prefix: str) -> str:
    return value.removeprefix(prefix)


def removesuffix(value: str, suffix: str) -> str:
    return value.removesuffix(suffix)


camelcase = stringcase.camelcase
spinalcase = stringcase.spinalcase
snakecase = stringcase.snakecase
pascalcase = stringcase.pascalcase


def lowercase(value: str) -> str:
    return value.lower()


def uppercase(value: str) -> str:
    return value.upper()



__all__ = [
    "PrincipalType",
    "UserId",
    "IAMArn",
]
import re
from typing import ClassVar, Optional, Pattern

from aibs_informatics_core.collections import StrEnum, ValidatedStr
from aibs_informatics_core.utils.decorators import cached_property


class PrincipalType(StrEnum):
    Account = "Account"
    User = "User"
    FederatedUser = "FederatedUser"
    AssumedRole = "AssumedRole"
    Anonymous = "Anonymous"


class UserId(ValidatedStr):
    """User ID returned from STS Caller Indentity

    Resources:
    - https://awscli.amazonaws.com/v2/documentation/api/latest/reference/sts/get-caller-identity.html
    - https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_variables.html#principaltable
    - https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_identifiers.html#identifiers-unique-ids

    """

    regex_pattern: ClassVar[Pattern] = re.compile(
        r"(?:anonymous|([\d]{12})|([\w\d]{19,21}))(?::(.+))?"
    )

    @property
    def account_id(self) -> Optional[str]:
        return self.get_match_groups()[0]

    @property
    def unique_id(self) -> Optional[str]:
        return self.get_match_groups()[1]

    @property
    def caller_specified_name(self) -> Optional[str]:
        return self.get_match_groups()[2]

    @cached_property
    def principal_type(self) -> PrincipalType:
        if self.account_id and not self.caller_specified_name:
            return PrincipalType.Account
        elif self.unique_id and not self.caller_specified_name:
            return PrincipalType.User
        elif self.account_id and self.caller_specified_name:
            return PrincipalType.FederatedUser
        elif self.unique_id and self.caller_specified_name:
            return PrincipalType.AssumedRole
        else:
            # self == "anonymous"
            return PrincipalType.Anonymous


# https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_identifiers.html
class IAMArn(ValidatedStr):
    """

    Possible ARN templates (variables in {}):
        arn:aws:iam::{account}:root
        arn:aws:iam::{account}:user/{user-name-with-path}
        arn:aws:iam::{account}:group/{group-name-with-path}
        arn:aws:iam::{account}:role/{role-name-with-path}
        arn:aws:iam::{account}:policy/{policy-name-with-path}
        arn:aws:iam::{account}:instance-profile/{instance-profile-name-with-path}
        arn:aws:sts::{account}:federated-user/{user-name}
        arn:aws:sts::{account}:assumed-role/{role-name}/{role-session-name}
        arn:aws:iam::{account}:mfa/{virtual-device-name-with-path}
        arn:aws:iam::{account}:u2f/{u2f-token-id}
        arn:aws:iam::{account}:server-certificate/{certificate-name-with-path}
        arn:aws:iam::{account}:saml-provider/{provider-name}
        arn:aws:iam::{account}:oidc-provider/{provider-name}

    Resources:
    - https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_identifiers.html
    """

    regex_pattern: ClassVar[Pattern] = re.compile(
        r"""arn:
            aws:
            (iam|sts):
            :                   # usually region, but IAM/STS resources are global
            ([\d]*):            # Account ID

            (
                (
                    root|assumed-role|user|group|role|policy|instance-profile|federated-user|server-certificate|mfa|u2f|.*-provider
                )
                (?:/(.+))?
            )
        """,
        re.X,
    )

    @property
    def account_id(self) -> str:
        return self.get_match_groups()[1]

    @property
    def resource(self) -> str:
        """The resource of the ARN

        Examples:
            - root
            - user/JohnDoe
            - assumed-role/Accounting-Role/JaneDoe
        Returns:
            str: resource
        """
        return self.get_match_groups()[2]

    @property
    def resource_type(self) -> str:
        """The resource Type of the ARN

        Examples are root, user, assumed-role

        Returns:
            str: AWS resource type
        """
        return self.get_match_groups()[3]

    @property
    def resource_id(self) -> str:
        """The resource ID of the ARN

        For most, this follows the first forward slash, but for "root", we return the account id

        Returns:
            str: AWS resource type
        """
        return self.get_match_groups()[4] or self.account_id

    @property
    def resource_name(self) -> str:
        return self.resource_id.split("/")[-1]

    @property
    def resource_path(self) -> str:
        return f"/{'/'.join(self.resource_id.split('/')[:-1])}"



from dataclasses import dataclass
from typing import Optional

from aibs_informatics_core.models.base import SchemaModel, custom_field


@dataclass
class AWSBatchExecutionPlatform(SchemaModel):
    job_queue_name: str


# TODO: I would prefer to make ExecutionPlatform polymorphic, but datacalasses does not support it
#       For now, I will just make a
@dataclass
class ExecutionPlatform(SchemaModel):
    aws_batch: Optional[AWSBatchExecutionPlatform] = custom_field(
        mm_field=AWSBatchExecutionPlatform.as_mm_field(), default=None
    )
